//------------------------------------------------------------------------------------------------------------------------
// Class       : BP_EventLink
// Author      : BehemothProgrammer
// Description : Triggers actors/script when condition is met also when restored unless flagged not to.
//               Chain multiple EventLinks together for more conditions.
//               One main use for EventLinks is to keep enemies that were activated and alive on the map still active next time the level is loaded.
// OnTrigger   : If condition is met then it will trigger target actors and/or call scriptID.
//
// Parameters  : (7) Params 1 -128..127					: Condition
//               (6) Params 2 -128..127					: Condition Arg 1
//               (4) Attach Chance 0..255				: Condition Arg 2
//               Height									: Condition Arg 3
//               (3) Max Regenerations -32768..32767	: Flags
//               (0) Target TID -32768..32767 			: Actors TID to trigger. If is less than 1 this will not be used.
//               (2) Health -32768..32767				: Script to trigger. If is less than 1 this will not be used.
//               Radius									: Delay before triggering actors/script
//               WallRadius								: Amount of time to wait until can be triggered again (If is repeatable)
//               StepHeight/DeadHeight					: 
//------------------------------------------------------------------------------------------------------------------------

enum BPEventLinkFlags
{
    BP_ELF_REPEATABLE = (1 << 0), //Triggers every OnActivate when condition is met instead of once. Is never marked persistent.
    BP_ELF_INVERSECONDITION = (1 << 1),
    BP_ELF_NOPERSISTCALL = (1 << 2) //Will not trigger on map start when marked persistent
}

enum BPEventLinkConditions
{
    BP_ELC_NONE = 0,
    BP_ELC_DIFFICULTY = 1, //Arg1 is the difficulty flags that must be on (BPEventLinkConditionDifficultyFlags)
    BP_ELC_PLAYER_GROUNDED = 2, //Player must be grounded
    BP_ELC_INVENTORY = 3, //Has Inventory Item. Arg1 is the item index. Arg2 is the amount required to have.
    BP_ELC_KEY = 4, //Has collected key. Arg1 is the hubID (normally the first keys are for hub 2). Arg2 is the keybit.
    BP_ELC_ACTOR_ACTIVATED = 5, //Actor is activated. Arg3 is the TID of the actor.
	BP_ELC_ACTOR_DEAD = 6 //Actor is Dead or doesn't exist. Arg3 is the TID of the actor.
}

enum BPEventLinkConditionDifficultyFlags
{
    BP_ELC_DF_EASY = (1 << 0),
    BP_ELC_DF_NORMAL = (1 << 1),
    BP_ELC_DF_HARD = (1 << 2),
    BP_ELC_DF_HARDCORE = (1 << 3)
}

enum BPEventLinkStates
{
    BP_ELS_IDLE = 0,
	BP_ELS_DELAYING,
	BP_ELS_RETRIGGERDELAY,
}

class BP_EventLink : BP_ScriptObject
{
	int condition;
	int conditionArg1;
	int conditionArg2;
	int conditionArg3;
	int flags;
	int triggerTID;
	int triggerScriptID;
	float delay;
	float retriggerDelay;
	
	kActor@ conditionActor;
	//------------------------------------------------------------------------------------------------------------------------
    BP_EventLink(kActor @actor)
	{
		super(actor);
		state = BP_ELS_IDLE;
    }
	//------------------------------------------------------------------------------------------------------------------------
    void OnBeginLevel()
    {
		condition = self.SpawnParams(7);
		conditionArg1 = self.SpawnParams(6);
		conditionArg2 = self.SpawnParams(4);
		conditionArg3 = int(self.Height());
		flags = self.SpawnParams(3);
		triggerTID = self.SpawnParams(0);
		triggerScriptID = self.SpawnParams(2);
		delay = self.Radius();
		retriggerDelay = self.WallRadius();
		
		switch (condition)
		{
			case BP_ELC_ACTOR_ACTIVATED:
			case BP_ELC_ACTOR_DEAD:
				@conditionActor = World.GetActorByTID(conditionArg3);
				break;
		}
		
		self.Flags() &= ~AF_SOLID;
		self.Flags() |= AF_INVINCIBLE;
	}
	//------------------------------------------------------------------------------------------------------------------------
    void OnRestore()
    {
		self.Flags() |= AF_ACTIVATED;
    }
	//------------------------------------------------------------------------------------------------------------------------
	void OnPostBeginLevel()
	{
		if (self.IsPersistentMarked() && flags & BP_ELF_NOPERSISTCALL == 0)
			TriggerLink();
	}
	//------------------------------------------------------------------------------------------------------------------------
	void OnTick() override
	{
		switch (state)
		{
			case BP_ELS_IDLE:
			{
				break;
			}
			case BP_ELS_DELAYING:
			{
				if (Time::Reached(stateEndTime))
					TriggerLink();
				break;
			}
			case BP_ELS_RETRIGGERDELAY:
			{
				if (Time::Reached(stateEndTime))
					TriggerLink();
				break;
			}
		}
	}
	//------------------------------------------------------------------------------------------------------------------------
	void OnActivate()
	{
		if (!IsConditionMet())
		{
			self.Flags() &= ~AF_ACTIVATED;
			return;
		}
		
		ScheduleTriggerLink();
	}
	//------------------------------------------------------------------------------------------------------------------------
	void OnStateChange() override
	{
		switch (state)
		{
			case BP_ELS_IDLE:
			{
				self.Flags() &= ~AF_ALWAYSACTIVE;
				if (flags & BP_ELF_REPEATABLE != 0)
					self.Flags() &= ~AF_ACTIVATED;
				break;
			}
			case BP_ELS_DELAYING:
			{
				self.Flags() |= AF_ALWAYSACTIVE;
				self.Flags() &= ~AF_DORMANT;
				stateEndTime = Time::FromNow(delay);
				break;
			}
			case BP_ELS_RETRIGGERDELAY:
			{
				self.Flags() |= AF_ALWAYSACTIVE;
				self.Flags() &= ~AF_DORMANT;
				stateEndTime = Time::FromNow(retriggerDelay);
				break;
			}
		}
	}
	//------------------------------------------------------------------------------------------------------------------------
	void ScheduleTriggerLink()
	{
		if (flags & BP_ELF_REPEATABLE == 0)
			self.MarkPersistentBit(false);
		
		if (delay <= 0.0f)
		{
			TriggerLink();
			return;
		}
		
		SetState(BP_ELS_DELAYING);
	}
	//------------------------------------------------------------------------------------------------------------------------
	void TriggerLink()
	{
		//Log("EventLink " + self.TID() + " Triggered TID " + triggerTID + " and Script " + triggerScriptID);
		
		if (retriggerDelay <= 0.0f || flags & BP_ELF_REPEATABLE == 0)
			SetState(BP_ELS_IDLE);
		else
			SetState(BP_ELS_RETRIGGERDELAY);
		
		if (flags & BP_ELF_REPEATABLE == 0)
			self.MarkPersistentBit(false);
		
		if (triggerTID > 0 && triggerTID != self.TID()) //uses actor TID and not triggering self again
			World.TriggerActorsByTID(self, triggerTID);
		if (triggerScriptID > 0)
			Game.CallDelayedMapScript(triggerScriptID, self, 0.0f);
	}
	//------------------------------------------------------------------------------------------------------------------------
	bool IsConditionMet()
	{
		bool result = true;
		switch (condition)
		{
			case BP_ELC_DIFFICULTY:
			{
				int difficulty = Game.GetDifficulty();
				result = ((difficulty == DIFFICULTY_EASY && conditionArg1 & BP_ELC_DF_EASY != 0) ||
						(difficulty == DIFFICULTY_NORMAL && conditionArg1 & BP_ELC_DF_NORMAL != 0) ||
						(difficulty == DIFFICULTY_HARD && conditionArg1 & BP_ELC_DF_HARD != 0) ||
						(difficulty == DIFFICULTY_HARDCORE && conditionArg1 & BP_ELC_DF_HARDCORE != 0));
				break;
			}
			case BP_ELC_PLAYER_GROUNDED:
			{
				result = Player.Actor().OnGround();
				break;
			}
			case BP_ELC_INVENTORY:
			{
				result = Inventory::Count(conditionArg1) >= conditionArg2;
				break;
			}
			case BP_ELC_KEY:
			{
				result = Inventory::HasKey(conditionArg1, conditionArg2);
				break;
			}
			case BP_ELC_ACTOR_ACTIVATED:
			{
				result = Actor::Exists(conditionActor) && Actor::IsActivated(conditionActor);
				break;
			}
			case BP_ELC_ACTOR_DEAD:
			{
				result = !Actor::Exists(conditionActor) || Actor::IsDead(conditionActor);
				break;
			}
		}
		
		return (flags & BP_ELF_INVERSECONDITION != 0) ? !result : result;
	}
	//------------------------------------------------------------------------------------------------------------------------
};
