//------------------------------------------------------------------------------------------------------------------------
// Class       : BP_SectorModifier
// Author      : BehemothProgrammer
// Description : Can modify sector area flags, or the 6 args, or water/sector height
// OnTrigger   : Starts modification of sector. If return delay is 0 then sets persistent bit.
//
// Parameters  : (7) Params 1 -128..127 				: Sector Modifier Type
//               (6) Params 2 -128..127 				: AreaFlagType - Area Flag Bit to turn on (1..32 negative values turn that bit off)
//														: AreaArgType - The arg to set. 1..6
//               Radius									: Start Delay
//               WallRadius								: Return delay (0 = no reset)
//               Height 								: Target value (AreaArg(-32768..32767), Sector/Water Height)
//               StepHeight/DeadHeight					: Time to reach target value (AreaArg, Sector/Water Height)
//               (3) Max Regenerations -32768..32767	: Flags
//               (0) Target TID -32768..32767 			: (Sound ID lookup. 526 is standard pillar move sound.)
//               (2) Health -32768..32767				: (Target Yaw Offset in degrees)
//               Spawn Flags 1 - Screen Shake			: Will shake the screen when moves
//------------------------------------------------------------------------------------------------------------------------

enum BPSectorModifierTypes
{
    BP_SM_TYPE_AREAFLAG = 0,
    BP_SM_TYPE_AREAARG = 1,
    BP_SM_TYPE_SECTORHEIGHT = 2, //also moves actor
    BP_SM_TYPE_WATERHEIGHT = 3, //also moves actor
    BP_SM_TYPE_AREATELEPORTARGS = 4 //Changes Area Arg1 and 2. Height is arg1 (TID) and Heath becomes arg2 (MapID). Param 2 - AreaArgType is not used.
}

enum BPSectorModifierFlags
{
	BP_SM_FLAG_PERPETUAL = (1 << 0), //After return has finished will start moving again. Will also always return even with delay 0
	BP_SM_FLAG_PLAYERGROUNDED = (1 << 1), //Triggers only when the player is grounded
	BP_SM_FLAG_NOACTORMOVE = (1 << 2), //Will not move this actor up/down for Sector or Water Modifier Types
	BP_SM_FLAG_PLAYERROTATE = (1 << 3) //If the player is grounded and on the area then the player will rotate too
}

enum BPSectorModifierStates
{
    BP_SMS_NONE = 0,
    BP_SMS_STARTIDLE = 1,
	BP_SMS_STARTDELAY = 2,
	BP_SMS_STARTING = 3,
	BP_SMS_RETURNIDLE = 4,
	BP_SMS_RETURNDELAY = 5,
	BP_SMS_RETURNING = 6
}

class BP_SectorModifier : BP_ScriptObject
{
	int modifierType;
	int areaFlagBit;
	uint areaFlags;
	int areaArg;
	int targetArgValue;
	int targetArgValue2;
	int flags;
	int moveSoundID;
	float targetYawOffset;
	float startDelay;
	float returnDelay;
	float targetHeight;
	float targetTime;
	
	kVec3 startOrigin;
	int startSector;
	int startArea;
	array<int> startAreaArgs;
	float startActorZ;
	float startSectorHeight;
	float startWaterHeight;
	float startYaw;
	//------------------------------------------------------------------------------------------------------------------------
    BP_SectorModifier(kActor @actor)
    {
		super(actor);
		state = BP_SMS_NONE;
    }
	//------------------------------------------------------------------------------------------------------------------------
    void OnBeginLevel()
    {
		modifierType = self.SpawnParams(7);
		areaFlagBit = Math::Clamp(self.SpawnParams(6), -32, 32);
		areaFlags = uint(1 << (areaFlagBit - 1));
		areaArg = Math::Clamp(self.SpawnParams(6), 1, 6) - 1;
		flags = self.SpawnParams(3);
		moveSoundID = self.SpawnParams(0);
		targetYawOffset = Math::Deg2Rad(self.SpawnParams(2));
		targetArgValue2 = self.SpawnParams(2);
		startDelay = self.Radius();
		returnDelay = self.WallRadius();
		targetHeight = self.Height();
		targetArgValue = int16(self.Height());
		targetTime = self.StepHeight();
		
		startOrigin = self.Origin();
		startArea = self.AreaID();
		startSector = self.SectorIndex();
		for (int i = 0; i < 6; i++)
			startAreaArgs.insertLast(World.GetAreaArg(startArea, i));
		startActorZ = self.Origin().z;
		startYaw = self.Yaw();
		startSectorHeight = self.FloorHeight();
		startWaterHeight = self.GetWaterHeight();
		
		self.Flags() &= ~AF_SOLID;
		self.Flags() |= AF_INVINCIBLE;
		
		if (self.IsPersistentMarked())
		{
			if (flags & BP_SM_FLAG_PERPETUAL != 0)
			{
				if (startDelay > 0.0f)
					SetState(BP_SMS_STARTDELAY);
				else
					SetState(BP_SMS_STARTING);
			}
			else
			{
				SetState(BP_SMS_RETURNIDLE);
			}
		}
		else
		{
			SetState(BP_SMS_STARTIDLE);
		}
    }
	//------------------------------------------------------------------------------------------------------------------------
	void OnTick() override
	{
		TickState();
	}
	//------------------------------------------------------------------------------------------------------------------------
    void OnActivate()
    {
		self.Flags() &= ~AF_ACTIVATED;
		
		if (flags & BP_SM_FLAG_PLAYERGROUNDED != 0 && !Player.Actor().OnGround())
			return;
		
		if (state != BP_SMS_STARTIDLE)
			return;
		
		if (returnDelay <= 0.0f)
			self.MarkPersistentBit(false);
		
		if (startDelay > 0.0f)
			SetState(BP_SMS_STARTDELAY);
		else
			SetState(BP_SMS_STARTING);
    }
	//------------------------------------------------------------------------------------------------------------------------
	void OnStateChange() override
	{
		stateStartTime = Time::FromNow(0.0f);
		switch (state)
		{
			case BP_SMS_STARTIDLE:
			{
				ModifySector(0.0f);
				break;
			}
			case BP_SMS_STARTDELAY:
			{
				ModifySector(0.0f);
				stateEndTime = Time::FromNow(startDelay);
				break;
			}
			case BP_SMS_STARTING:
			{
				PlayMoveSound();
				MoveShake();
				ModifySector(0.0f);
				stateEndTime = Time::FromNow(targetTime);
				break;
			}
			case BP_SMS_RETURNIDLE:
			{
				ModifySector(1.0f);
				break;
			}
			case BP_SMS_RETURNDELAY:
			{
				ModifySector(1.0f);
				stateEndTime = Time::FromNow(returnDelay);
				break;
			}
			case BP_SMS_RETURNING:
			{
				PlayMoveSound();
				MoveShake();
				ModifySector(1.0f);
				stateEndTime = Time::FromNow(targetTime);
				break;
			}
		}
	}
	//------------------------------------------------------------------------------------------------------------------------
	void TickState()
	{
		switch (state)
		{
			case BP_SMS_STARTIDLE:
			{
				break;
			}
			case BP_SMS_STARTDELAY:
			{
				if (Time::Reached(stateEndTime))
					SetState(BP_SMS_STARTING);
				break;
			}
			case BP_SMS_STARTING:
			{
				float normTime = Time::Normalize(stateStartTime, stateEndTime);
				ModifySector(normTime);
				if (normTime >= 1.0f)
				{
					self.StopSound();
					if (returnDelay > 0.0f)
						SetState(BP_SMS_RETURNDELAY);
					else if (flags & BP_SM_FLAG_PERPETUAL != 0)
						SetState(BP_SMS_RETURNING);
					else
						SetState(BP_SMS_RETURNIDLE);
				}
				break;
			}
			case BP_SMS_RETURNIDLE:
			{
				break;
			}
			case BP_SMS_RETURNDELAY:
			{
				if (Time::Reached(stateEndTime))
					SetState(BP_SMS_RETURNING);
				break;
			}
			case BP_SMS_RETURNING:
			{
				float normTime = Time::Normalize(stateStartTime, stateEndTime);
				ModifySector(1.0f - normTime);
				if (normTime >= 1.0f)
				{
					self.StopSound();
					if (flags & BP_SM_FLAG_PERPETUAL != 0)
					{
						if (startDelay > 0.0f)
							SetState(BP_SMS_STARTDELAY);
						else
							SetState(BP_SMS_STARTING);
					}
					else
					{
						SetState(BP_SMS_STARTIDLE);
					}
				}
				break;
			}
		}
	}
	//------------------------------------------------------------------------------------------------------------------------
	void ModifySector(const float normTime)
	{
		if (targetYawOffset != 0.0f && modifierType != BP_SM_TYPE_AREATELEPORTARGS)
		{
			kAngle lastYaw = self.Yaw();
			self.Yaw() = Math::SmoothStepFloat(startYaw, startYaw + targetYawOffset, normTime); 
			if (flags & BP_SM_FLAG_PLAYERROTATE != 0)
			{
				kActor@ pActor = Player.Actor().CastToActor();
				if (pActor.AreaID() == startArea && pActor.OnGround())
					pActor.Yaw() += self.Yaw().Diff(lastYaw);
			}
		}
		
		switch (modifierType)
		{
			case BP_SM_TYPE_AREAFLAG:
			{
				if (areaFlagBit != 0 && (normTime >= 1.0f || normTime <= 0.0f))
				{
					World.FloodMatchingAreaFlags(startSector, areaFlags, normTime >= 1.0f);
				}
				break;
			}
			case BP_SM_TYPE_AREAARG:
			{
				World.ChangeAreaArg(startArea, areaArg, Math::LerpInt(startAreaArgs[areaArg], targetArgValue, normTime));
				break;
			}
			case BP_SM_TYPE_SECTORHEIGHT:
			{
				if (flags & BP_SM_FLAG_NOACTORMOVE == 0)
					self.Origin().z = Math::SmoothStepFloat(startActorZ, startActorZ + (targetHeight - startSectorHeight), normTime);
				World.ChangeSectorHeight(startSector, Math::SmoothStepFloat(startSectorHeight, targetHeight, normTime));
				break;
			}
			case BP_SM_TYPE_WATERHEIGHT:
			{
				if (flags & BP_SM_FLAG_NOACTORMOVE == 0)
					self.Origin().z = Math::SmoothStepFloat(startActorZ, startActorZ + (targetHeight - startWaterHeight), normTime);
				World.ChangeAreaWaterHeight(startArea, Math::SmoothStepFloat(startWaterHeight, targetHeight, normTime));
				break;
			}
			case BP_SM_TYPE_AREATELEPORTARGS:
			{
				World.ChangeAreaArg(startArea, BP_AREAARG_WARP_TID, Math::LerpInt(startAreaArgs[BP_AREAARG_WARP_TID], targetArgValue, normTime));
				World.ChangeAreaArg(startArea, BP_AREAARG_WARP_MAPID, Math::LerpInt(startAreaArgs[BP_AREAARG_WARP_MAPID], targetArgValue2, normTime));
				break;
			}
		}
	}
	//------------------------------------------------------------------------------------------------------------------------
	void PlayMoveSound()
	{
		if (moveSoundID > 0)
			self.PlaySoundWithLookupID(moveSoundID);
	}
	//------------------------------------------------------------------------------------------------------------------------
	void MoveShake()
	{
		//has screen shake spawn flag on
		if ((self.SpawnFlags1() & (1 << 30)) != 0)
			Game::AddShake(self, targetTime);
	}
	//------------------------------------------------------------------------------------------------------------------------
};
