//------------------------------------------------------------------------------------------------------------------------
// Class       : BP_Destructible
// Author      : BehemothProgrammer
// Description : Once this actor is killed it will play it's destructibleDeath animation and
//               can trigger other actors or map scripts. And can drop an item with a percent chance.
// OnTrigger   : Nothing
//
// Parameters  : (3) Max Regenerations -32768..32767	: Flags
//               WallRadius								: ActorID to spawn when destroyed (Item Drop) If is less than 1 this will not be used.
//               (6) Params 2 -128..127					: 0..100% chance of spawning the item drop.
//               (0) Target TID -32768..32767 			: Actors TID to trigger when destroyed. If is less than 1 this will not be used.
//               (2) Health -32768..32767				: Script to trigger when destroyed. If is less than 1 this will not be used.
//
// Extra Defs  : BP_DestructibleFlags					: Sets flags in .ctor (for spawned actors)
//               BP_DestructibleItemChance				: Sets itemChance in .ctor (for spawned actors)
//------------------------------------------------------------------------------------------------------------------------

enum BPDestructibleFlags
{
    BP_DESTRUCTIBLE_FLAG_ONLY_EXPLOSION_IMPACTS = (1 << 0),
    BP_DESTRUCTIBLE_FLAG_ONLY_NONEXPLOSION_IMPACTS = (1 << 1),
    BP_DESTRUCTIBLE_FLAG_FLASH_ONDAMAGE = (1 << 2),
    BP_DESTRUCTIBLE_FLAG_NODRAW_ONSLEEP = (1 << 3)
}

class BP_Destructible : BP_ScriptObject
{
	int flags;
	int itemChance;
	int triggerTID;
	int triggerScriptID;
	int health;
	//------------------------------------------------------------------------------------------------------------------------
    BP_Destructible(kActor @actor)
    {
		super(actor);
		itemChance = 100;
		flags = Actor::GetDefInt(self, DEF_DESTRUCTIBLE_FLAGS);
		itemChance = Actor::GetDefInt(self, DEF_DESTRUCTIBLE_ITEMCHANCE, itemChance);
		
		//make dead so OnDeath will not get called (for spawned actors)
		health = self.Health();
		self.Health() = 0;
    }
	//------------------------------------------------------------------------------------------------------------------------
    void OnBeginLevel()
    {
		flags = self.SpawnParams(3);
		triggerTID = self.SpawnParams(0);
		triggerScriptID = self.SpawnParams(2);
		itemChance = Math::Clamp(self.SpawnParams(6), 0, 100);
		
		//make dead so OnDeath will not get called (for map actors)
		health = self.Health();
		self.Health() = 0;
	}
	//------------------------------------------------------------------------------------------------------------------------
    void OnRestore()
    {
        self.Flags() |= AF_DEAD | AF_IGNORESOUNDEVENTS | AF_INVINCIBLE;
        self.Flags() &= ~AF_SOLID;
		self.Radius() = 0.0f;
		self.Height() = 0.0f;
        self.AnimState().Set(anim_destructibleDeath, 4.0f, 0);
        self.AnimState().SetLastFrame(true);
    }
	//------------------------------------------------------------------------------------------------------------------------
	void OnSleep()
	{
		if (flags & BP_DESTRUCTIBLE_FLAG_NODRAW_ONSLEEP != 0)
			self.Flags() |= AF_NODRAW;
	}
	//------------------------------------------------------------------------------------------------------------------------
	void OnWake()
	{
		if (flags & BP_DESTRUCTIBLE_FLAG_NODRAW_ONSLEEP != 0)
			self.Flags() &= ~AF_NODRAW;
	}
	//------------------------------------------------------------------------------------------------------------------------
    void OnDamage(kActor @instigator, kDictMem @damageDef, const int damage)
    {
		if (CanBeDamaged(instigator, damageDef))
		{
			health -= damage;
			if (health > 0)
			{
				if (flags & BP_DESTRUCTIBLE_FLAG_FLASH_ONDAMAGE != 0)
					self.RunFxEvent("Mantis_Hit");
			}
			else
			{
				//Do Death
				self.Flags() |= AF_DEAD | AF_INVINCIBLE;
				self.Flags() &= ~AF_SOLID;
				self.Radius() = 0.0f;
				self.Height() = 0.0f;
				self.AnimState().Blend(anim_destructibleDeath, 4.0f, 4.0f, 0);
				self.MarkPersistentBit(false);
				
				//Trigger Actors/Script
				if (triggerTID > 0)
					World.TriggerActorsByTID(self, triggerTID);
				if (triggerScriptID > 0)
					Game.CallDelayedMapScript(triggerScriptID, self, 0.0f);
				
				//Try spawning item drop
				int actorID = int(self.WallRadius());
				if (actorID > 0 && Math::RandMax(100) < itemChance)
				{
					kActor@ item = Actor::Spawn(actorID, self.Origin(), self.Yaw(), self.SectorIndex());
					if (Actor::Exists(item))
						Actor::TossItem(item);
				}
			}
		}
    }
	//------------------------------------------------------------------------------------------------------------------------
	bool CanBeDamaged(kActor@ instigator, kDictMem @damageDef)
	{
		bool explosiveDamage = false;
		if (damageDef !is null)
			damageDef.GetBool(DEF_DAMAGE_EXPLODE, explosiveDamage);
		
		if (flags & BP_DESTRUCTIBLE_FLAG_ONLY_EXPLOSION_IMPACTS != 0)
			return explosiveDamage;
		else if (flags & BP_DESTRUCTIBLE_FLAG_ONLY_NONEXPLOSION_IMPACTS != 0)
			return !explosiveDamage;
		
		return true;
	}
	//------------------------------------------------------------------------------------------------------------------------
    void MeleeBluntHeavy(kActor @instigator, const float w, const float x, const float y, const float z)
    {
        self.MeleeObject("Damage_Blunt_20", kVec3(x, y, z), w);
    }
	//------------------------------------------------------------------------------------------------------------------------
};
